"""Custom search function to assist in customizing our special coverage searches."""
from collections import OrderedDict

from django.conf import settings

from elasticsearch_dsl import filter as es_filter

from bulbs.content.custom_search import get_condition_filter
from bulbs.content.models import Content


class SearchParty(object):
    """Wrapper that searches for content from a group of SpecialCoverage objects."""

    def __init__(self, special_coverages):
        """Store list of special_coverages class level."""
        self._special_coverages = special_coverages
        self._query = {}

    def search(self):
        """Return a search using the combined query of all associated special coverage objects."""
        # Retrieve all Or filters pertinent to the special coverage query.
        should_filters = [
            es_filter.Terms(pk=self.query.get("included_ids", [])),
            es_filter.Terms(pk=self.query.get("pinned_ids", []))
        ]
        should_filters += self.get_group_filters()

        # Compile list of all Must filters.
        must_filters = [
            es_filter.Bool(should=should_filters),
            ~es_filter.Terms(pk=self.query.get("excluded_ids", []))
        ]

        return Content.search_objects.search().filter(es_filter.Bool(must=must_filters))

    def get_group_filters(self):
        """Return es OR filters to include all special coverage group conditions."""
        group_filters = []
        field_map = {
            "feature-type": "feature_type.slug",
            "tag": "tags.slug",
            "content-type": "_type"
        }
        for group_set in self.query.get("groups", []):
            for group in group_set:
                group_filter = es_filter.MatchAll()
                for condition in group.get("conditions", []):
                    group_filter &= get_condition_filter(condition, field_map=field_map)
                group_filters.append(group_filter)
        return group_filters

    @property
    def query(self):
        """Group the self.special_coverages queries and memoize them."""
        if not self._query:
            self._query.update({
                "excluded_ids": [],
                "included_ids": [],
                "pinned_ids": [],
                "groups": [],
            })
            for special_coverage in self._special_coverages:
                # Access query at dict level.
                query = getattr(special_coverage, "query", {})
                if "query" in query:
                    query = query.get("query")
                self._query["excluded_ids"] += query.get("excluded_ids", [])
                self._query["included_ids"] += query.get("included_ids", [])
                self._query["pinned_ids"] += query.get("pinned_ids", [])
                self._query["groups"] += [query.get("groups", [])]
        return self._query


class SearchSlicer(object):
    """We want to search things like a seesaw. take that mike parent.
    """
    def __init__(self, *args, **kwargs):
        self.default_queryset = None
        self.querysets = OrderedDict()
        self.index = 0
        self.limit = kwargs.get("limit", getattr(settings, "READING_LIST_LIMIT", 100))

    def __iter__(self):
        return self

    def next(self):
        if self.index >= self.limit:
            raise StopIteration
        for validator, queryset in self.querysets.items():
            if validator(self.index):
                try:
                    result = queryset.next()
                    self.index += 1
                    return result
                except StopIteration:
                    pass
        result = self.default_queryset.next()
        self.index += 1
        return result

    def __next__(self):
        return self.next()

    def register_queryset(self, queryset, validator=None, default=False):
        """
        Add a given queryset to the iterator with custom logic for iteration.

        :param queryset: List of objects included in the reading list.
        :param validator: Custom logic to determine a queryset's position in a reading_list.
            Validators must accept an index as an argument and return a truthy value.
        :param default: Sets the given queryset as the primary queryset when no validator applies.
        """
        if default or self.default_queryset is None:
            self.default_queryset = queryset
            return
        if validator:
            self.querysets[validator] = queryset
        else:
            raise ValueError(
                """Querysets require validation logic to integrate with reading lists."""
            )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('content', '0004_auto_20151103_1358'),
        ('contributions', '0004_auto_20151103_1358'),
    ]

    operations = [
        migrations.CreateModel(
            name='ReportContent',
            fields=[
            ],
            options={
                'proxy': True,
            },
            bases=('content.content',),
        ),
        migrations.AddField(
            model_name='freelanceprofile',
            name='payroll_name',
            field=models.CharField(max_length=256, null=True, blank=True),
        ),
    ]



"""ElasticSearch filters specifially for content"""

import datetime
import dateutil.parser
import dateutil.tz
from django.utils import timezone
from elasticsearch_dsl.filter import Term, Terms, Range, MatchAll, Nested
from six import string_types, text_type, binary_type


def Evergreen(evergreen=True):
    return Term(evergreen=evergreen)


def parse_datetime(value):
    """Returns a datetime object for a given argument

    This helps to convert strings, dates and datetimes to proper tz-enabled
    datetime objects."""

    if isinstance(value, (string_types, text_type, binary_type)):
        value = dateutil.parser.parse(value)
        value.replace(tzinfo=dateutil.tz.tzutc())
        return value
    elif isinstance(value, datetime.datetime):
        value.replace(tzinfo=dateutil.tz.tzutc())
        return value
    elif isinstance(value, datetime.date):
        value = datetime.datetime(value.year, value.month, value.day)
        value.replace(tzinfo=dateutil.tz.tzutc())
        return value
    else:
        raise ValueError('Value must be parsable to datetime object. Got `{}`'.format(type(value)))


def Published(before=None, after=None):  # noqa
    published_params = {}
    if after is not None:
        published_params["gte"] = parse_datetime(after)

    if before is not None:
        published_params["lte"] = parse_datetime(before)

    if before is None and after is None:
        published_params["lte"] = timezone.now()

    return Range(published=published_params)


def Status(status):  # noqa
    if status:
        return Term(status=status)
    else:
        return MatchAll()


def Authors(usernames):  # noqa
    included = []
    excluded = []
    for username in usernames:
        if username.startswith("-"):
            excluded.append(username[1:])
        else:
            included.append(username)
    f = MatchAll()
    if included:
        f &= Terms(**{"authors.username": included})
    if excluded:
        f &= Terms(**{"authors.username": excluded})
    return f


def Tags(slugs):  # noqa
    included = []
    excluded = []
    for slug in slugs:
        if slug.startswith("-"):
            excluded.append(slug[1:])
        else:
            included.append(slug)

    f = MatchAll()
    if included:
        f &= Nested(path="tags", filter=Terms(**{"tags.slug": included}))
    if excluded:
        f &= ~Nested(path="tags", filter=Terms(**{"tags.slug": excluded}))

    return f


def FeatureTypes(slugs):  # noqa
    included = []
    excluded = []
    for slug in slugs:
        if slug.startswith("-"):
            excluded.append(slug[1:])
        else:
            included.append(slug)

    f = MatchAll()
    if included:
        f &= Nested(path="feature_type", filter=Terms(**{"feature_type.slug": included}))
    if excluded:
        f &= ~Nested(path="feature_type", filter=Terms(**{"feature_type.slug": excluded}))

    return f



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contributions', '0006_auto_20160105_1502'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='featuretyperate',
            options={'ordering': ('feature_type__name',)},
        ),
        migrations.AlterModelOptions(
            name='lineitem',
            options={'ordering': ('-payment_date',)},
        ),
    ]



from rest_framework import filters, routers, viewsets
from rest_framework.permissions import IsAdminUser

from .models import Section
from .serializers import SectionSerializer


class SectionViewSet(viewsets.ModelViewSet):
    queryset = Section.objects.all()
    serializer_class = SectionSerializer
    filter_backends = (filters.SearchFilter, filters.OrderingFilter,)
    search_fields = ("name",)
    ordering_fields = ("name",)
    paginate_by = 10
    permission_classes = [IsAdminUser]

api_v1_router = routers.DefaultRouter()
api_v1_router.register(
    r"section",
    SectionViewSet,
    base_name="section")



from django.conf.urls import url, patterns

urlpatterns = patterns("bulbs.redirects.views",
    url(r"^(?P<pk>\d+)/?$", "utm_redirect", name="utm-redirect-pk"),   # noqa
    url(r"^(?P<pk>\d+)(?P<source>[a-z])(?P<medium>[a-z])(?P<name>[a-z])/?$",
        "utm_redirect",
        name="utm-redirect-tracking")
)



from django.conf import settings
from django.db import models
from django.db.models.signals import pre_delete
from django.template.defaultfilters import slugify

from bulbs.content.custom_search import custom_search_model
from bulbs.content.models import Content, ElasticsearchImageField
from djes.models import Indexable
from elasticsearch import Elasticsearch
from elasticsearch_dsl import field
from djbetty import ImageField
from json_field import JSONField


es = Elasticsearch(settings.ES_CONNECTIONS["default"]["hosts"])


class Section(Indexable):

    name = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(max_length=255, blank=True, editable=True, unique=True)
    description = models.TextField(default="", blank=True)
    embed_code = models.TextField(default="", blank=True)
    section_logo = ImageField(null=True, blank=True)
    twitter_handle = models.CharField(max_length=255, blank=True)
    promoted = models.BooleanField(default=False)
    query = JSONField(default={}, blank=True)

    class Mapping:
        name = field.String(analyzer="autocomplete", fields={"raw": field.String(index="not_analyzed")})
        slug = field.String(index="not_analyzed")
        section_logo = ElasticsearchImageField()
        query = field.Object(enabled=False)

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        """Saving ensures that the slug, if not set, is set to the slugified name."""

        if not self.slug:
            self.slug = slugify(self.name)

        section = super(Section, self).save(*args, **kwargs)

        if self.query and self.query != {}:
            self._save_percolator()

        return section

    def _save_percolator(self):
        """saves the query field as an elasticsearch percolator
        """
        index = Content.search_objects.mapping.index
        query_filter = self.get_content().to_dict()

        q = {}

        if "query" in query_filter:
            q = {"query": query_filter.get("query", {})}
        else:
            return

        es.index(
            index=index,
            doc_type=".percolator",
            body=q,
            id=self.es_id
        )

    def _delete_percolator(self):
        index = Content.search_objects.mapping.index
        es.delete(index=index, doc_type=".percolator", id=self.es_id, refresh=True, ignore=404)

    def get_content(self):
        """performs es search and gets content objects
        """
        if "query" in self.query:
            q = self.query["query"]
        else:
            q = self.query
        search = custom_search_model(Content, q, field_map={
            "feature-type": "feature_type.slug",
            "tag": "tags.slug",
            "content-type": "_type",
            })
        return search

    @property
    def es_id(self):
        return "section.{}".format(self.id)

    @property
    def contents(self):
        """Caches the results of the get_content method
        """
        if not hasattr(self, "_content"):
            self._content = self.get_content()
        return self._content


def remove_percolator(sender, instance, *args, **kwargs):
    instance._delete_percolator()

pre_delete.connect(remove_percolator, sender=Section)



from rest_framework import serializers
from rest_framework.utils import model_meta

from djbetty.serializers import ImageFieldSerializer

from .models import Campaign, CampaignPixel


class PixelTypeField(serializers.Field):
    """
    Pixel 'type' objects serialized to/from label/identifier
    """
    def to_representation(self, obj):
        return dict(CampaignPixel.PIXEL_TYPES)[obj]

    def to_internal_value(self, data):
        if isinstance(data, int):
            return data
        return dict((label, value)
                for value, label in CampaignPixel.PIXEL_TYPES)[data]


class CampaignPixelField(serializers.ModelSerializer):

    pixel_type = PixelTypeField()

    class Meta:
        model = CampaignPixel
        exclude = ("campaign",)

    def save(self, campaign=None):
        self.instance.campaign = campaign
        return super(CampaignPixelField, self).save()


class CampaignSerializer(serializers.ModelSerializer):

    sponsor_logo = ImageFieldSerializer(required=False)
    pixels = CampaignPixelField(many=True)
    start_date = serializers.DateTimeField()
    end_date = serializers.DateTimeField()

    class Meta:
        model = Campaign

    def create(self, validated_data):
        ModelClass = self.Meta.model

        # Remove many-to-many relationships from validated_data.
        # They are not valid arguments to the default `.create()` method,
        # as they require that the instance has already been saved.
        info = model_meta.get_field_info(ModelClass)
        many_to_many = {}
        for field_name, relation_info in info.relations.items():
            if relation_info.to_many and (field_name in validated_data):
                many_to_many[field_name] = validated_data.pop(field_name)

        instance = ModelClass.objects.create(**validated_data)

        # Save many-to-many relationships after the instance is created.
        if many_to_many:

            pixel_serializer = CampaignPixelField(data=many_to_many["pixels"], many=True)
            if not pixel_serializer.is_valid():
                raise Exception("Ivalid pixel data")
            pixel_serializer.save(campaign=instance)

        return instance

    def update(self, instance, validated_data):

        pixels_data = validated_data.pop("pixels")

        for attr, value in validated_data.items():
            setattr(instance, attr, value)
        instance.save()

        if pixels_data:
            existing_pixel_ids = list(instance.pixels.values_list("id", flat=True))
            for pixel_data in pixels_data:
                if pixel_data.get("id") in existing_pixel_ids:
                    existing_pixel_ids.remove(pixel_data["id"])

            pixel_serializer = CampaignPixelField(data=pixels_data, many=True)
            if not pixel_serializer.is_valid():
                raise Exception("Ivalid pixel data")
            pixel_serializer.save(campaign=instance)

            for pixel_id in existing_pixel_ids:
                print(pixel_id)
                # We need to remove these...
                CampaignPixel.objects.get(id=pixel_id).delete()
        else:
            for pixel in instance.pixels.all():
                pixel.delete()

        return instance



from django.http import Http404
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_control


from bulbs.special_coverage.models import SpecialCoverage
from bulbs.content.views import BaseContentDetailView


class SpecialCoverageView(BaseContentDetailView):
    redirect_correct_path = False

    def get_template_names(self):
        template_names = ["special_coverage/landing.html"]
        return template_names

    def get_object(self, *args, **kwargs):
        self.special_coverage = get_object_or_404(SpecialCoverage, slug=self.kwargs.get("slug"))

        if not self.special_coverage.is_active:
            raise Http404("Special Coverage does not exist.")

        qs = self.special_coverage.get_content(published=self.show_published_only()).full()
        if qs.count() == 0:
            raise Http404("No Content available in content list")

        return qs[0]

    def get_context_data(self, *args, **kwargs):
        context = super(SpecialCoverageView, self).get_context_data()
        context["content_list"] = self.special_coverage.get_content()
        context["special_coverage"] = self.special_coverage

        return context


special_coverage = cache_control(max_age=600)(SpecialCoverageView.as_view())
